#charset "us-ascii"

/* Copyright (c) 2000, 2002 Michael J. Roberts.  All Rights Reserved. */
/*
 *   TADS 3 Library - Status Line
 *   
 *   This module defines the framework for displaying the status line,
 *   which is the area conventionally displayed at the top of the screen
 *   showing information such as the current location, score (if scoring
 *   is used at all), and number of turns.  
 */

/* include the library header */
#include "adv3.h"


/* ------------------------------------------------------------------------ */
/*
 *   The OutputStream for the status line.  We use a separate output
 *   stream for this so that we don't affect the main text area when
 *   updating the status line.  
 */
statuslineOutputStream: OutputStream
    /*
     *   The statusline is currently handled by writing to the main system
     *   console, so we must treat it as sharing the keyboard for input.
     *   At some point in the future, we envision a new console output API
     *   that will explicitly provide a separate system output stream for
     *   the statusline, avoiding the need to coordinate with the input
     *   manager. 
     */
    myInputManager = inputManager

    /* we sit atop the system-level main console output stream */
    writeFromStream(txt)
    {
        /* write the text to the console */
        tadsSay(txt);
    }
;

/*
 *   The OutputStream for the right half of the status line (the
 *   score/turn count area) in text mode.  We use a separate stream for
 *   this because we have to write this text with the special
 *   statusRight() intrinsic in text mode. 
 */
statusRightOutputStream: OutputStream
    /*
     *   Write from the stream.  We simply buffer up text until we're
     *   asked to display the final data. 
     */
    writeFromStream(txt)
    {
        /* buffer the text */
        buf_ += txt;
    }

    /*
     *   Flush the buffer.  This writes whatever we've buffered up to the
     *   right half of the text-mode status line. 
     */
    flushStream()
    {
        /* write the text to the system console */
        statusRight(buf_);

        /* we no longer have anything buffered */
        buf_ = '';
    }

    /* our buffered text */
    buf_ = ''
;


/* ------------------------------------------------------------------------ */
/*
 *   Status line - this is an abstract object that controls the status
 *   line display.
 *   
 *   The object has two main methods: showStatusHtml, which shows the
 *   status line in HTML format, and showStatusText, which shows the
 *   status line in plain text mode.  The status line daemon calls one or
 *   the other of these methods, according to the current mode, to display
 *   the statusline.  The default implementations of these methods
 *   generate the appropriate formatting codes for a statusline with a
 *   left part and a right part, calling showStatusLeft and
 *   showStatusRight, respectively, to display the text for the parts.
 *   
 *   Games can customize the statusline at two levels.  At the simpler
 *   level, a game can modify showStatusLeft and/or showStatusRight to
 *   change the text displayed on the left and/or right of the statusline.
 *   Since these two methods are common to the text and HTML versions,
 *   games don't have to worry about the different modes when overriding
 *   these (unless they want to display their own HTML, of course, in
 *   which case they'll have to check or set the HTML mode).
 *   
 *   At the more difficult level, a game can modify showStatusHtml and/or
 *   showStatusText.  Modifying these routines provides complete control
 *   over the formatting.  If a game wants to use something other than the
 *   traditional left/right display, it must modify these methods.
 */
statusLine: object
    /* 
     *   Show the status line, in HTML or text mode, as appropriate.  By
     *   default, the library sets this up as a "prompt daemon," which
     *   means that this will be called automatically just before each
     *   command line is read.  
     */
    showStatusLine()
    {
        /* 
         *   Generate an HTML or plain text status line, depending on the
         *   current mode.  Generate HTML only if we're running in a fully
         *   HTML-enabled interpreter and we're currently in HTML mode.  
         */
        if (systemInfo(SysInfoHtml) && systemInfo(SysInfoHtmlMode))
        {
            /* show the HTML status line */
            statusLine.showStatusHtml();
        }
        else
        {
            /* show the status line in plain text mode */
            statusLine.showStatusText();
        }
    }

    /*
     *   Show the status line in HTML format.  Our default implementation
     *   shows the traditional two-part (left/right) status line, using
     *   showStatusLeft() and showStatusRight() to display the parts.  
     */
    showStatusHtml()
    {
        local oldStr;

        /* switch to the status line output stream */
        oldStr = outputManager.setOutputStream(statuslineOutputStream);

        /* make sure we restore statusline mode before we're done */
        try
        {
            /* write the statusline text in HTML mode */
            "<banner id=StatusLine height=previous border><body
            bgcolor=statusbg text=statustext><a plain
            href='<<libMessages.commandLookAround>>'>";
            
            /* show the left part of the status line */
            showStatusLeft();
            
            /* set up for the score part on the right half */
            "</a><tab align=right><a plain
            href='<<libMessages.commandFullScore>>'>";
            
            /* show the right part of the status line */
            showStatusRight();

            /* end the score link */
            "</a>";

            /* add the status-line exit list, if desired */
            gPlayerChar.location.showStatuslineExits();
            
            /* end the formatting */
            "</banner>";
        }
        finally
        {
            /* restore the old output stream */
            outputManager.setOutputStream(oldStr);
        }
    }

    /*
     *   Show the statusline in text mode.  Our default implementation
     *   shows the traditional two-part (left/right) status line, using
     *   showStatusLeft() and showStatusRight() to display the parts.  
     */
    showStatusText()
    {
        local oldStr;

        /* switch to the status line output stream */
        oldStr = outputManager.setOutputStream(statuslineOutputStream);

        /* enter statusline mode in the display layer */
        statusMode(StatModeStatus);

        /* make sure we exit statusline mode before we're finished */
        try
        {
            /* show the left part of the display */
            showStatusLeft();

            /* switch to the right-side status stream */
            outputManager.setOutputStream(statusRightOutputStream);

            /* show the right-half text */
            showStatusRight();

            /* flush the right-side stream */
            statusRightOutputStream.flushStream();
        }
        finally
        {
            /* exit status mode */
            statusMode(StatModeNormal);

            /* restore the old output stream */
            outputManager.setOutputStream(oldStr);
        }
    }

    /*
     *   Show the left part of a standard left/right statusline.  By
     *   default, we'll show the player character's 
     */
    showStatusLeft()
    {
        local actor;

        /* get the player character actor */
        actor = gPlayerChar;

        "<.statusroom>";

        /* show the actor's location's status name */
        if (actor.location != nil)
            actor.location.statusName(actor);

        "<./statusroom>";
    }

    /*
     *   Show the right part of a standard left/right statusline.  By
     *   default, we'll show the current score, a slash, and the number of
     *   turns. 
     */
    showStatusRight()
    {
        local s;

        /* if there's a score object, show the score */
        if ((s = libGlobal.scoreObj) != nil)
        {
            /* show the score and the number of turns so far */
            "<.statusscore><<s.totalScore>>/<<
            libGlobal.totalTurns>><./statusscore>";
        }
    }
;

/* in case the 'score' module isn't included */
property totalScore;
